#include "stdlib.h"
#include "malloc.h"


struct node {
	int value;
	struct node *next;
};

/*@

predicate node(struct node *no, int v, struct node *ne)
	= no->value |-> v &*& no->next |-> ne &*& no != 0;

@*/

struct llist {
	struct node *head;
};

/*@

inductive intlist = | my_nil | my_cons(int, intlist);

predicate llist(struct llist *n, intlist a)
	= n->head |-> ?np &*& list(np, a);
	
predicate list(struct node *n, intlist a)
	= n == 0 ? a == my_nil
	         : node(n, ?i, ?ne) &*& list(ne, ?as) &*& a == my_cons(i, as);
	         
@*/

/*@

fixpoint intlist app(intlist l1, intlist l2) {
  switch (l1) {
    case my_nil: return l2;
    case my_cons(x, v): return my_cons(x, app(v, l2));
  }
}

fixpoint intlist rev(intlist l) {
  switch (l) {
    case my_nil: return my_nil;
    case my_cons(x, v): return app(rev(v), my_cons(x, my_nil));
  }
}

lemma void app_assoc(intlist l1, intlist l2, intlist l3)
  requires emp;
  ensures app(app(l1, l2), l3) == app(l1, app(l2, l3));
{
  switch (l1) {
    case my_nil:
    case my_cons(x, v):
      app_assoc(v, l2, l3);
  }
}

lemma void rev_lemma(int i, intlist l1, intlist l2)
  requires emp;
  ensures app(rev(my_cons(i, l1)), l2) == app(rev(l1), my_cons(i, l2));
{
  app_assoc(rev(l1), my_cons(i, my_nil), l2);
}

lemma void app_to_my_nil(intlist l)
  requires emp;
  ensures app(l, my_nil) == l;
{
  switch (l) {
    case my_nil:
    case my_cons(x, v):
      app_to_my_nil(v);
  }
}

lemma void rev_app(intlist l1, intlist l2)
  requires emp;
  ensures rev(app(l1, l2)) == app(rev(l2), rev(l1));
{
  switch (l1) {
    case my_nil:
      app_to_my_nil(rev(l2));
    case my_cons(x, v):
      rev_app(v, l2);
      app_assoc(rev(l2), rev(v), my_cons(x, my_nil));
  }
}

lemma void rev_twice(intlist l)
  requires emp;
  ensures rev(rev(l)) == l;
{
  switch (l) {
    case my_nil:
    case my_cons(x, v):
      rev_app(rev(v), my_cons(x, my_nil));
      rev_twice(v);
  }
}

@*/

void reverse(struct llist *l)
	//@ requires llist(l, ?a);
	//@ ensures llist(l, rev(a));
{
	struct node *previous = 0;
	//@ open llist(l, a);
	struct node *next = l->head;
	struct node *temp = 0;
	l->head = 0;
	//@ close list(0, my_nil);
	while (next != 0)
		//@ invariant l->head |-> ?n &*& list(n, ?b) &*& list(next, ?g) &*& a == app(rev(b), g);
	{
		previous = l->head;
		l->head = next;
		temp = next;
		//@ open list(temp, g);
		//@ open node(temp, ?i, _);
		next = temp->next;
		temp->next = previous;
		// The following two lines get a handle on the tail of g:
		//@ open list(next, ?gs);
		//@ close list(next, gs);
		//@ close node(temp, i, previous);
		//@ close list(temp, my_cons(i, b));
		//@ rev_lemma(i, b, gs);
	}
	//@ open list(next, _);
	//@ app_to_my_nil(rev(b));
	//@ rev_twice(b);
	//@ close llist(l, rev(a));
}



// Now, let's see if we can append a value to a linked list at the end.
// First, a helper function and a recursive algorithm.

struct node *new_node(int v, struct node *next)
	//@ requires emp;
	//@ ensures node(result, v, next) &*& result != 0;
{
	struct node *temp;
	temp = malloc(sizeof(struct node));
	if (temp == 0) abort();
	temp->value = v;
	temp->next = next;
	return temp;
	//@ close node(_, _, _);
	//@ leak malloc_block_node(temp);   // Leaking permission to free the node again...
}

struct node *recursive_append_helper(struct node *n, int v)
	//@ requires list(n, ?a);
	//@ ensures list(result, app(a, my_cons(v, my_nil))) &*& result != 0;
{
	struct node *temp;
	//@ open list(n, _);
	if (n == 0)
	{
		temp = new_node(v, 0);
		//@ close list(0, _);
		//@ close list(temp, _);
		return temp;
	}
	else
	{
		//@ open node(n, _, _);
		temp = recursive_append_helper(n->next, v);
		n->next = temp;
		//@ close node(n, _, _);
		//@ close list(n, _);
		return n;
	}
}

void recursive_append(struct llist *l, int v)
	//@ requires llist(l, ?a);
	//@ ensures llist(l, app(a, my_cons(v, my_nil)));
{
	struct node *temp;
	//@ open llist(l, _);
	temp = recursive_append_helper(l->head, v);
	l->head = temp;
	//@ close llist(l, _);
}


// Then an iterative algorithm.

/*@

predicate lseg(struct node *start, intlist l, struct node *end)
	= start == end ? emp &*& l == my_nil
		       : node(start, ?v, ?ne) &*& lseg(ne, ?vs, end) &*& l == my_cons(v, vs);

@*/

/*@

lemma void distinct_nodes(struct node *n1, struct node *n2)
	requires node(n1, ?v1, ?next1) &*& node(n2, ?v2, ?next2);
	ensures node(n1, v1, next1) &*& node(n2, v2, next2) &*& n1 != n2;
{
	open node(n1, _, _);
	open node(n2, _, _);
	close node(n1, _, _);
	close node(n2, _, _);
}

lemma void absorb_node(struct node *stop)
	requires lseg(?start, ?l, stop) &*& node(stop, ?v, ?final) &*& node(final, ?w, ?x);
	ensures lseg(start, app(l, my_cons(v, my_nil)), final) &*& node(final, w, x);
{
	distinct_nodes(stop, final);
	open lseg(start, _, _);
	if (start == stop) {
		close lseg(final, _, final);
	} else {
		distinct_nodes(start, final);
		open node(start, _, _);
		absorb_node(stop);
		close node(start, _, _);
	}
	close lseg(start, _, final);
}

lemma void lseg_to_list(struct node *end)
	requires lseg(?start, ?a, end) &*& node(end, ?v, 0);
	ensures list(start, app(a, my_cons(v, my_nil)));
{
	open lseg(_, _, _);
	if (start == end) {
		close list(0, _);
	} else {
		lseg_to_list(end);
	}
	open node(start, _, _);   // Just to say that start != 0.
	close node(start, _, _);
	close list(start, _);
}


@*/

void append(struct llist *l, int v)
	//@ requires llist(l, ?a);
	//@ ensures llist(l, app(a, my_cons(v, my_nil)));
{
	struct node *temp, temp2;
	//@ open llist(l, _);
	temp = l->head;
	if (temp == 0)
	{
		//@ open list(temp, _);
		temp2 = new_node(v, 0);
		//@ close list(0, _);
		//@ close list(temp2, _);
		l->head = temp2;
		//@ close llist(l, _);
	}
	else
	{
		//@ struct node *t = temp;
		//@ open list(temp, _);
		//@ open node(temp, _, _);
		//@ close lseg(t, _, temp);
		while (temp->next != 0)
			//@ invariant lseg(t, ?b, temp) &*& temp->value |-> ?w &*& temp->next |-> ?ne &*& temp!=0 &*& list(ne, ?c) &*& a == app(b, my_cons(w, c));
		{
			//@ struct node *i = temp;
			temp = temp->next;
			//@ close node(i, _, temp);
			//@ open list(temp, _);
			//@ absorb_node(i);
			//@ open node(temp, _, _);
			//@ app_assoc(b, my_cons(w, my_nil), c);
		}
		//@ open list(ne, _);
		temp2 = new_node(v, 0);
		temp->next = temp2;
		//@ close node(temp, _, _);
		//@ absorb_node(temp);
		//@ lseg_to_list(temp2);
		//@ close llist(l, _);
	}
}
